package rvl.swv;
import rvl.swv.*;
import java.io.*;

/**
  This is an abstract implementation of <tt>EngineInterface</tt>
  that can be extended to easily create an <tt>EngineInterface</tt>
  for a new piece of software.
  <p> To use, extend this class and define these
      variables in the constructor (defaults, if any, are shown):
  <ul>
    <li> {@link #engineName}
    <li> {@link #codeExt} and {@link #outExt}
    <li> {@link #commentStart} and {@link #commentEnd}
    <li> {@link #startupCode} and {@link #endCode}
    <li> {@link #okfigfmts}
  </ul>
  In the configuration file, define <i>engineName</i><tt>.binary</tt>
  to a string giving the command to run the program.  It should contain the
  substring <tt>%codefile%</tt>, and optionally, <tt>%outfile%</tt>.
  The appropriate filenames are substituted to form the command line that is executed.

  <p>In most cases, only a few methods need to be overridden:
  <ul>
  <li> {@link #putExpr}
  <li> {@link #putSeparator}
  <li> {@link #setupFig}<br>
  <li> {@link #closeFig}
  </ul>
 */


public abstract class AbstractEngine implements EngineInterface {

// These are the things that should be defined in the constructor...

/** Name of this engine -- used in error msgs and such */
    public String engineName;

/** Extension to add to the file's base name for the code file
    (include the initial "dot")
    This should be unique to this engine. */
    public String codeExt;
/** Extension to add to the file's base name for the output file
    produced by running the code file (include the initial "dot")
    This should be unique to this engine, if possible */
    public String outExt;

/** String to put before a comment in the code file.  Default is <tt>"/* "</tt> */
    public String commentStart = "/* ";
/** String to put after a comment in the code file. Default is <tt>" *</tt><tt>/"</tt> */
    public String commentEnd = " */";

/** Code to put at beginning of code file, before any code chunks */
    public String startupCode="";
/** Code to put at beginning of code file, before any code chunks */
    public String endCode="";

/** List of figure file formats supported by this engine */
    public int okfigfmts[] = new int[] {FigFile.PNG, FigFile.JPG};


// Other variables that likely don't need to be messed with.
    String baseName, codeFileName, outFileName, binary;
    String pathSep = System.getProperty("file.separator");

    File codeFile;

    BufferedWriter oStream;

    StatWeave parent;

    FileInterface filei;

    boolean weaving;  // set by putStartupCode()



    /** Default constructor */
    public AbstractEngine() {
    }

    /** Open a file for saving the code. Return <tt>true</tt> if successful */
    public boolean openCodeFile(String baseName) {
        try {
            this.baseName = baseName;
            codeFileName = baseName + codeExt;
            outFileName = baseName + outExt;
            codeFile = new File(codeFileName);
            oStream = new BufferedWriter(
                new FileWriter(codeFile));
            return true;
        }
        catch (IOException e) {
            return false;
        }
    }

    /** Close the code file */
    public void closeCodeFile() {
        if (weaving) {
            putCode(endCode);
        }
        try {
            oStream.close();
        }
        catch (IOException e) {
            parent.warning(engineName + ": Error in closing code file");
        }
    }

    /** Delete the code and output files */
    public void deleteCodeFile() {
        codeFile.delete();
    }

    /** Return code and output file names */
    public String[] getFileNames() {
        return new String[] {codeFileName, outFileName};
    }

    /** Set the parents that drive this instance */
    public void setParents(StatWeave parent, FileInterface filei) {
        this.parent = parent;
        this.filei = filei;
    }

    /** Set where to find the program to run on the code.
    (This need not be called if only tangling)
    */
    public void setBinary(String binloc) {
        binary = binloc;
    }

    /** Add a line of code to the code file */
    public void putCode(String line) {
        try {
            oStream.write(line + "\n");
        }
        catch (IOException e) {
            parent.error(engineName + ": Error in writing to code file", 20);
        }
    }

    /** Evaluate an expression and put it in the output stream */
    public void putExpr(String expr) {
        putCode(expr);
    }



    /** Startup code */
    public void putStartup(boolean weaving) {
        this.weaving = weaving;
        putCode(commentStart + " File generated by StatWeave " + commentEnd + "\n");
        putSeparator("--- Output file from StatWeave, engine = " + engineName + " ---");
        if (weaving) {
            putCode(startupCode);
        }
    }

    /** Put code that will add the given text in a separate line of
    the engine's output. This is used by the parent to keep track of
    beginnings and ends of output chunks.
    */
    public void putSeparator(String text) {
        putCode("put \"" + text + "\"");
    }

    /** Add a comment line to the source code */
    public void putComment(String text) {
        putCode(commentStart + text + commentEnd);
    }

    /**
    Put code that will set up a file to receive figure output
    This method uses the contents of {@link #okfigfmts} and negotiates
    with the current {@link FileInterface} to determine what
    figure format to actually use, and how to size it.
    The extending class should call this as a superclass method,
    then use the properties of the returned object as needed in
    outputting the needed setup code for the figure
    @return <tt>null</tt> if no format can be found; otherwise the
    object can be used by the extending class to finish setting up the figure.
    */
    public FigFile setupFig(String chunkName) {
        int fmts[] = filei.getFigFormats(),
            fmt = FigFile.NULL;
        for (int i=0; fmt==FigFile.NULL && i<fmts.length; i++) {
            for (int j=0; j<okfigfmts.length; j++)
                if (fmts[i] == okfigfmts[j] && fmt==FigFile.NULL)
                    fmt = fmts[i];
        }
        double wid = parent.getDim("width",6), ht = parent.getDim("height",6);
        if (fmt == FigFile.NULL) {
            parent.warning("Warning in " + engineName +
                ": No compatible graphics formats");
            return null;
        }
        String lbl = FigFile.labels[fmt],
            fn = chunkName + "-fig." + lbl;
        return new FigFile(fn, fmt, wid, ht, 0);
    }

    /** Put code to close the graphics file created by <TT>setupFig</TT> */
    public void closeFig() {
    }

    /** Run the code, and return its <TT>Process</TT> object */
    public Process runCode() {
        try {
            String cmd = binary.replaceAll("%codefile%", codeFileName);
            cmd = cmd.replaceAll("%outfile%", outFileName);
            parent.message("\n-----\n" + cmd);
            return Runtime.getRuntime().exec(cmd);
        }
        catch (IOException ioe) {
            ioe.printStackTrace(parent.errStream);
            parent.error("Runtime error in " + engineName
                + " (possible configuration issue)", Err.SYSERR);
        }
        return null;
    }

}

