package rvl.swv;
import rvl.swv.*;

/**
  This interface specifies the methods needed to read source files
  and write result files for StatWeave.
 */
public interface FileInterface {


    /** Set the parent object for this instance */
    public void setParent(StatWeave parent);

    /** Access the parent */
    public StatWeave getParent();

    /**
    Return acceptable graphics formats -- null if no graphics supported.<p>
    Note the required integer constants are defined in {@link FigFile}
    */
    public int[] getFigFormats();


    /**
    Read the specified source file (most likely into a buffer).
    Implementations are responsible for closing this file.
    */
    public void readSourceFile(String fileName);

    /**
    Return the next line from the source file.  May be needed by a SyntaxInterface
    when options continue to the next line.  If no SyntaxInterface is used
    just have this return null.
    */
    public String getLine();

    /**
    Return a string that describes somehow the current position in the source file
    */
    public String getPosition();

    /**
    Write the results to the results file
    (a call to <tt>parent.getConfig("sys.baseName")</tt> may help figure out the right name.)
    <p>Finally, clean up intermediate files as directed by the <tt>mopUp</tt>
    argument, which should have a value of 0, 1, or 2.  0 means leave all intermediate
    files; 2 means erase all intermediate files and leave only essential ones; and
    1 is somewhere in between (e.g. keep graphics files) */
    public void writeResults(int mopUp);

    /** Return the next tag type found in the file */
    public Tag nextTag();

    /** Return the language associated with latest tag
    <p> Needed when tag is <tt>chunk</tt>, <tt>langOpts</tt>, or <tt>expr</tt>
    */
    public String getLang();

    /** Return the expression given in current expr tag,
    or the label in a <tt>recallCode</tt>, <tt>recallOutput</tt>, or <tt>recallFig</tt> tag */
    public String getExpr();

    /** Return the options list for latest tag.
    <p> Needed when tag is <tt>chunk</tt>, <tt>globalOpts</tt>, or <tt>langOpts</tt>
    */
    public String getOptions();

    /** Return one line from a code chunk.
    <p>If it is actually a code reference, return the string
    <tt>!coderef <i>label</i></tt> or <tt>!coderef *<i>label</i></tt>,
    where <tt><i>label</i></tt> is the label provided.  Use the <tt>*</tt>
    form to force the code reference to be displayed regardless of
    the value of the option <tt>showref</tt>.
    Note that in either case, the label can be obtained using <tt>.substring(9)</tt>.
    <p>Return <tt>null</tt> when the end of the code chunk is reached.
    */
    public String getCodeLine();

    /** Return the label specified in the latest tag.
    <p>Needed when the tag is <tt>coderef</tt>, <tt>recallFig</tt>, <tt>recallCode</tt>,
    <tt>recallOutput</tt> or when <tt>getCodeLine</tt> returns a code reference.
    (When the tag is <tt>chunk</tt>, the label, if any, should be included
    in the list returned by <tt>getOptions()</tt>.) */
    public String getLabel();

    /** Return the arguments specified in a code reference.. */
    public java.util.Vector<String> getArgs();

    /** Replace the current tagged region of the source file
    with the specified <tt>sig</tt>.  This provides a place-holder for results to be
    incorporated later. */
    public void saveBookmark(String sig);

    /** Find the given bookmark previously saved with signature <tt>sig</tt>, and replace it
    with text.  The <tt>context</tt> argument is a reminder of what kind of content
    is being added.  Implementation may call <tt>StatWeave</tt>'s <tt>getOption()</tt> method
    and expect to retrieve the options in effect for the text being inserted. */
    public void replaceBookmark(String sig, String text, Tag context);

    /** Replace bookmark with specified figure(s) in <tt>ff</tt> */
    public void embedPlots(String sig, FigFile ff, int[] pages);


}
