package rvl.swv;
import rvl.swv.*;
import java.io.*;
import java.util.*;

/**
  This class defines the <tt>FileInterface</tt> for a LaTeX file.<p>
  The following format-specific options are implemented:
  <ul>
  <li> <tt>codestyle = </tt>styleName - environment name for code listing
  <li> <tt>outstyle = </tt>styleName - environment name for verbatim output listing
  <li> <tt>codefmt</tt> - custom verbatim formats for code block
  <li> <tt>outfmt</tt> - custom verbatim formats for output block
  <li> <tt>results</tt> - disables verbatim output when equal to <tt>tex</tt>
  <li> <tt>beforecode</tt>, <tt>aftercode</tt> - additional LaTeX markup to put before and after code listings
  <li> <tt>beforeout</tt>, <tt>afterout</tt> - additional LaTeX markup to put before and after output listings
  <li> <tt>beforefig</tt>, <tt>afterfig</tt> - additional LaTeX markup to put before and after figures
  </ul>
 */
public class LaTeXFile implements FileInterface {

    protected StatWeave parent;      // StatWeave instance that called this

    protected SyntaxInterface syni;  // The syntax interface we'll use

    protected BufferedReader iStream;    // source file stream
    String thisLine, thisLang, thisExpr;       // Most recently read line and language
    Tag thisContext;                 // Context of most recent tag
    int lineNo, thisLinePos;         // Pointer inside <tt>thisLine</tt>
    boolean inlineInProgress = false; // Flag -- true if still looking at thisLine

// This is our main working storage.  We accumjulate the source file into this
// and modify it as directed; then write it to the result file.
    protected StringBuilder storage;

    boolean pkgFound = false;  // Flag for finding "Statweave}" in preamble


/** Default constructor */
    public LaTeXFile() {
    }


/** Save the <tt>StatWeave</tt> instance that we are serving */
    public void setParent(StatWeave parent) {
        this.parent = parent;
    }

/** Access the <tt>StatWeave</tt> instance that we are serving */
    public StatWeave getParent() {
        return parent;
    }

/**
Return acceptable graphics formats -- null if no graphics supported.
<p>Note: most required integer constants are defined in {@link FigFile}
*/
    public int[] getFigFormats() {
        String target = parent.getConfig("sys.target"),
            figfmt = parent.getConfig("LaTeXFile.figfmt." + target);
        int fmt = FigFile.fmtID(figfmt);
        return (fmt == FigFile.PS || fmt == FigFile.EPS
            ? new int[] { FigFile.PS, FigFile.EPS }
            : new int[] {fmt, FigFile.PDF, FigFile.PNG, FigFile.JPG} );
    }



// New methods for my revised interface

    /**
    Read the specified source file.  For this impl., we also figure out which syntax
    (<tt>SyntaxInterface</tt>) to use.  We will just set up an input stream
    and read it line-by-line.  We need to close it later when all is done.
    */
    public void readSourceFile(String fileName) {
        try {
            iStream = new BufferedReader(
                new FileReader(fileName));
                storage = new StringBuilder();
        }
        catch (FileNotFoundException e) {
            parent.error("Source file \"" + fileName + "\" not found", Err.FILENF);
        }
      // Figure out the required syntax
        String synint = parent.getConfig("LaTeXFile.SyntaxInterfaces");
        if (synint == null) parent.error("Missing required key from configuration: LaTeXFile.SyntaxInterfaces", Err.NOKEY);
        StringTokenizer st = new StringTokenizer(synint);
        String ext = null, syn = "???";
        int dotIndex;
        while (ext == null && st.hasMoreTokens()) {
            syn = st.nextToken();
            ext = parent.match(fileName, parent.getConfig(syn + ".sources"));
        }
        if (ext == null) { // if no match, go back & get 1st one
            st = new StringTokenizer(parent.getConfig("LaTeXFile.SyntaxInterfaces"));
            syn = st.nextToken();
        }
        try {
            syni  = (SyntaxInterface)(Class.forName(parent.getConfig(syn + ".class"))).newInstance();
            syni.setFileInterface(this);
        }
        catch (Exception exc) {
            exc.printStackTrace(parent.errStream);
            parent.error("Unable to create needed interface(s)", Err.REFLECT);
        }
        lineNo = 0;
    }

    /**
    Write the results to the output file derived from sys.basename.
    <p>Finally, clean up intermediate files as directed by the <tt>mopUp</tt>
    argument, which should have a value of 0, 1, or 2.  0 means leave all intermediate
    files; 2 means erase all intermediate files and leave only essential ones; and
    1 is somewhere in between (e.g. keep graphics files) */
    public void writeResults(int mopUp) {
        String target = parent.getConfig("sys.target"),
            baseName = parent.getConfig("sys.basename"),
            texName = baseName + ".tex";
        try {
            BufferedWriter oStream = new BufferedWriter(
                new FileWriter(texName));
            oStream.write(storage.toString());
            oStream.close();
        }
        catch (IOException ioe) {
            ioe.printStackTrace(parent.errStream);
            parent.error("Unable to write output file \"" + texName + "\"", Err.IOERR);
        }
      // Are we done yet?  Maybe, but check target
        if (target.equals("pdf") || target.equals("dvi")) {
            String cmd = parent.getConfig("LaTeXFile.bin." + target);
            if (cmd == null)
                parent.error("Missing required configuration key LaTeXFile.bin." + target, Err.NOKEY);
            cmd += " " + texName;
            try {
                parent.message("\n-----\n" + cmd);
                Process proc = Runtime.getRuntime().exec(cmd);
                try {
                    // show output of process
                    BufferedReader prcOut = new BufferedReader(
                        new InputStreamReader(proc.getInputStream()));
                    for (String prcLine = prcOut.readLine(); prcLine != null; prcLine = prcOut.readLine())
                        parent.message(prcLine);
                }
                catch (IOException ioe) {
                    parent.error("Unable to read process output", Err.IOERR);
                }
                int rtncode = proc.waitFor();
                if (rtncode > 0)
                    parent.error("Problem in making " + target, Err.CHILDERR);
            }
            catch (IOException ioe) {
                ioe.printStackTrace(parent.errStream);
                parent.error("I/O error (possibly can't find binary)", Err.IOERR);
            }
            catch (InterruptedException e) {
                parent.error("Process was interrupted", Err.CHILDINT);
            }
        }
        parent.message("Successfully created " + baseName + "." + target);
    }

    /** Return the next tag type found in the file */
    public Tag nextTag() {

        while (true) { // infinite loop -- keep trying until something returns!

            if (inlineInProgress) { // Still looking for inline tags
                thisContext = syni.testInline(thisLine);
                if (thisContext != null) {
                    String det = syni.getDetails();
                    if (thisContext == Tag.expr) {
                        int p = det.indexOf(" ");
                        thisLang = det.substring(0,p);
                        thisExpr = det.substring(p+1).trim();
                    }
                    else {
                        thisExpr = det;
                    }
                    return thisContext;
                }

              // (else) Got here by finding no more inline stuff
                inlineInProgress = false; // Nothing found. Proceed to input a new line
            }

            getLine(); // this will also reset thisLine

            if (thisLine == null) { // No more tags -- so close the source file
                try {
                    iStream.close();
                }
                catch (IOException ioe2) {
                    ioe2.printStackTrace(parent.errStream);
                    parent.warning("LaTeXFile: Error in closing source file");
                }
                return Tag.noMore;
            }

          //else implied
            thisLang = syni.testChunk(thisLine);
            if (thisLang != null)
                return (thisContext = Tag.chunk);

         //else implied
            thisLang = syni.testOpts(thisLine);
            if (thisLang != null)
                return (thisContext = thisLang.equals("") ? Tag.globalOpts : Tag.langOpts);

         //got here because line does not begin a chunk or option spec
            storage.append(thisLine + "\n"); // add it to our storage
            inlineInProgress = true; // next time around, check for inline stuff

        } // infinite loop

    }

    /** Read one line from source file */
    public String getLine() {
        try {
            thisLine = iStream.readLine();
            if (thisLine == null)
                return null;
            thisLinePos = 0;
            lineNo++;
            if (thisLine.indexOf("usepackage{Statweave}") > 0) pkgFound = true;
            if (!pkgFound && thisLine.startsWith("\\begin{document}")) {
                thisLine = "\\usepackage{Statweave}\n" + thisLine;
                pkgFound = true;
            }
        }
        catch (IOException ioe) {
            ioe.printStackTrace(parent.errStream);
            parent.error("LaTeXFile: Error in reading source file in " + getPosition(), Err.SRCERR);
        }
        return thisLine;
    }

    /**
    * Return the current line number in the form <tt>"line #"</tt>.
    */
    public String getPosition() {
        return "line " + lineNo;
    }


 /** Return one line from a code chunk.
    <p>If it is actually a code reference, return the string
    <tt>!coderef <i>label</i></tt> or <tt>!coderef <i>*label</i></tt>,
    where <tt><i>label</i></tt> is the label provided.  (Use the <tt>*</tt>
    form to force the code reference to be displayed regardless of
    the value of the option <tt>showref</tt>.)
    <p>Return <tt>null</tt> when the end of the code chunk is reached.
    */
    public String getCodeLine() {
        getLine();
        if (syni.isChunkEnd(thisLine))
            return null;
        String lab = syni.testCoderef(thisLine);
        if (lab == null) return thisLine;
        return lab == null ? thisLine : "!coderef " + lab;
    }


    /** Return the language associated with latest tag
    <p> Needed when tag is <tt>chunk</tt>, <tt>langOpts</tt>, or <tt>expr</tt> */
    public String getLang() {
        return thisLang;
    }

    /** Return the expression given in the current expr tag */
    public String getExpr() {
        return thisExpr;
    }

    /** Return the options list for latest tag.
    <p> Needed when tag is <tt>chunk</tt>, <tt>globalOpts</tt>, <tt>langOpts</tt>,
    <tt>recallFig</tt>, <tt>recallCode</tt>, or <tt>recallOutput</tt>.
    */
    public String getOptions() {
        return syni.getOptionList();
    }

    /** Return the label specified in the latest tag.
    <p>Needed when the tag is <tt>coderef</tt>, <tt>recallFig</tt>, <tt>recallCode</tt>,
    <tt>recallOutput</tt>.  (When the tag is <tt>chunk</tt>, the label, if any, should be included
    in the list returned by <tt>getOptions()</tt>.) */
    public String getLabel() {
        return thisExpr;  // we're using thisExpr to save the label
    }

    /** Return the arguments specified in a <tt>coderef</tt> tag. */
    public java.util.Vector<String> getArgs() {
        return syni.getArgs();
    }

    /** Replace the current tagged region of the source file
    with the specified <tt>sig</tt>.  This provides a place-holder for results to be
    incorporated later. */
    public void saveBookmark(String sig) {
        if (thisContext == Tag.chunk || thisContext == Tag.coderef
                || thisContext == Tag.globalOpts || thisContext == Tag.langOpts)
            storage.append(sig);
        else {
            String pat = syni.getLastMatch();
            int pos = storage.indexOf(pat);
            if (pos < 0)
                parent.error("Internal error in saveBookmark()", Err.SYSERR);
            storage.delete(pos, pos + pat.length());
            storage.insert(pos, sig);
        }
    }

    /** Find the given bookmark previously saved with signature <tt>sig</tt>, and replace it
    with text.  The <tt>context</tt> argument is a reminder of what kind of content
    is being added.  Implementation may call <tt>StatWeave</tt>'s <tt>getOption()</tt> method
    and expect to retrieve the options in effect for the text being inserted. */
    public void replaceBookmark(String sig, String text, Tag context) {
        int p = storage.indexOf(sig);
        String bef=null, aft=null;
        if (p<0) {
            parent.warning("\nWarning: Bookmark not found.  Perhaps you recalled something that was never generated.");
            return;
        }
        storage.delete(p, p+sig.length());
        if (text.trim().equals("")) return;

        StringBuilder sb = new StringBuilder(text);

        if (context == Tag.recallOutput) {
            String res = parent.getOption("results");
            boolean verbatim = res==null ? true : res.equals("tex") ? false : true;
            if (verbatim) {
                String fmt = parent.getOption("outfmt"),
                    style = parent.getOption("outstyle", "Woutput");
              //append the stuff we need at beginning in REVERSE order
                sb.insert(0, "\n");
                if (fmt != null)
                    sb.insert(0, "[" + fmt + "]");
                sb.insert(0, "\\begin{" + style + "}");
              //close the environment
                sb.append("\\end{" + style + "}\n");
            }
            bef = parent.getOption("beforeout");
            aft = parent.getOption("afterout");
        }
        else if (context == Tag.recallCode) { // we're
            String fmt = parent.getOption("codefmt"),
                style = parent.getOption("codestyle", "Winput");
          //append the stuff we need at beginning in REVERSE order
            sb.insert(0, "\n");
            if (fmt != null)
                sb.insert(0, "[" + fmt + "]");
            sb.insert(0, "\\begin{" + style + "}");
          //close the environment
            sb.append("\\end{" + style + "}\n");
            bef = parent.getOption("beforecode");
            aft = parent.getOption("aftercode");
        }
        if (bef != null) sb.insert(0, bef + "\n");
        if (aft != null) sb.append(aft + "\n");
        storage.insert(p, sb.toString());
    }


    /** Replace bookmark with specified figure(s) in <tt>ff</tt> */
    public void embedPlots(String sig, FigFile ff, int[] pages) {
        int p = storage.indexOf(sig);
        if (p<0)
            parent.error("System error: bookmark " + sig + " not found", Err.SYSERR);
        storage.delete(p, p+sig.length());
        String opts = "",
            scale = parent.getOption("scale"),
            bef = parent.getOption("beforefig"),
            aft = parent.getOption("afterfig"),
            cmd;
        if (scale != null)
            opts = "scale=" + scale;
        else {
            double wid = parent.getDim("dispw",0), ht = parent.getDim("disph",0);
            if (wid > 0) opts = "width=" + wid + (ht > 0 ? "in," : "in");
            if (ht > 0) opts += "height=" + ht + "in";
        }
        cmd = (bef==null ? "" : bef + "\n") + "\\includegraphics[" + opts;
        for (int i=0; i<pages.length; i++) {
            if (pages[i] > 1) cmd += ", page=" + pages[i];
            cmd += "]{" + ff.getFileName() + "}\n";
        }
        if (aft != null) cmd += aft + "\n";
        storage.insert(p, cmd);
    }


} //----------- END ------------------------------------

