package rvl.swv;
import rvl.swv.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.util.regex.*;

/**
  This class defines the <tt>FileInterface</tt> for a ODT file.<p>
  The following format-specific options are implemented:
  <ul>
  <li> <tt>codestyle = </tt>styleName - custom paragraph style for code block
  <li> <tt>outstyle = </tt>styleName - custom paragraph style for output block
  <li> <tt>results</tt> - disable verbatim output when equal to <tt>xml</tt>
  </ul>
 */
public class ODTFile implements FileInterface {

    StatWeave parent;      // StatWeave instance that called this
    JarFile srcODT;        // JarFile object for source
    BufferedReader iStream;    // source file stream

    String inFile,
        thisLine, thisLang, lastLang="null", thisExpr, thisOpts;
        // Most recently read line, language, etc.
    String thisParTags;       // current  <text:p ...>  content
    String pendingPar = null;
    String lfSig = "%@-ODT-LF-@%"; // signature for spaces to put before and after chunks.
                                   // to be replaced by blank lines when writing output
    int currBeg=0, currEnd=0, prevEnd=0,  // current position in buffer
        exprBeg, exprEnd,    // positions WITHIN thisLine
        parNo=0;

    StringBuilder inbuffer, outbuffer; // Incoming and outgoing XML code

    Vector<String> codeStorage, codeRefArgs,  // Areas for saving code and args
         plotFiles;           // list of plot filenames
    Iterator<String> codeIter;

    Pattern spacePat, colonPat, bracePat;

    boolean exprFlag = false;  // flag -- true when expressions are pending

    // Strings for substituting back and forth between XML and code
    String xmlSub[][] = new String[][] {
        {"&gt;", ">"},
        {"&lt;", "<"},
        {"&quot;", "\""},
        {"&apos;", "'"},
        {"&amp;", "&"}};

    // Patterns for substituting only from XML to code, not back
    // This is crap generated by "autocorrect" mode
    String xmlSub1[][] = new String[][] {
        {"\u00c2\u00bd", "1/2"},
        {"\u00c2\u00bc", "1/4"},
        {"\u00c2\u00a9", "(c)"},
        {"\u00c2\u00ae", "(r)"},
        {"\u00e2\u20ac\u02dc", "'"},  // opening single quote
        {"\u00e2\u20ac\u2122", "'"},  // closing single quote
        {"\u00e2\u20ac\u0153", "\""},  // opening double quote
        {"\u00e2\u20ac\ufffd", "\""},  // closing double quote
        {"\u00e2\u20ac\u201c", "-"}};  // en dash


/** Default constructor */
    public ODTFile() {
        spacePat = Pattern.compile("<text:s text:c=\"([\\d]+)\"/>");
        colonPat = Pattern.compile("(^[a-zA-Z]+):");  // e.g. "SAS: ..."
        bracePat = Pattern.compile("\\{([^\\}]*)\\}");
        plotFiles = new Vector<String>();
    }


/** Save the <tt>StatWeave</tt> instance that we are serving */
    public void setParent(StatWeave parent) {
        this.parent = parent;
    }

/** Access the <tt>StatWeave</tt> instance that we are serving */
    public StatWeave getParent() {
        return parent;
    }

/**
Return acceptable graphics formats -- null if no graphics supported.
<p>Note: most required integer constants are defined in {@link FigFile}
*/
    public int[] getFigFormats() {
        return new int[] {FigFile.PNG, FigFile.JPG};
    }



// New methods for my revised interface

    /**
    Read the specified source file.  For this impl., we also figure out which syntax
    (<tt>SyntaxInterface</tt>) to use.  We will just set up an input stream
    and read it line-by-line.  We need to close it later when all is done.
    */
    public void readSourceFile(String fileName) {
        inFile = fileName;
        try {
            srcODT = new JarFile(fileName);
            JarEntry content = srcODT.getJarEntry("content.xml");
            iStream = new BufferedReader(
                new InputStreamReader(srcODT.getInputStream(content)));
            inbuffer = new StringBuilder();
            char buf[] = new char[2000];
            int count;
            while ((count = iStream.read(buf, 0, 2000)) != -1)
                inbuffer.append(buf, 0, count);
            outbuffer = new StringBuilder();
        }
        catch (FileNotFoundException e) {
            parent.error("Source file \"" + fileName + "\" not found", Err.FILENF);
        }
        catch (IOException ioe) {
            parent.error("Problem in reading \"" + fileName + "\"", Err.IOERR);
        }
    }

    /**
    Write the results to the output file derived from sys.basename.
    <p>Finally, clean up intermediate files as directed by the <tt>mopUp</tt>
    argument, which should have a value of 0, 1, or 2.  0 means leave all intermediate
    files; 2 means erase all intermediate files and leave only essential ones; and
    1 is somewhere in between (e.g. keep graphics files) */
    public void writeResults(int mopUp) {
        String baseName = parent.getConfig("sys.basename"),
            outFile = baseName + ".odt";
        if (outFile.equals(inFile)) {
            outFile = baseName + "-out.odt";
            parent.warning("Note: Result filename changed to " + outFile);
        }

        // fix up line-feed signatures...
        int lng = lfSig.length(), idx;
        while ((idx=outbuffer.indexOf(lfSig+lfSig)) > 0)  // remove consecutive lfs
            outbuffer.replace(idx, idx + lng, "");
        while ((idx=outbuffer.indexOf(lfSig)) > 0)  // replace lfs with blank lines
            outbuffer.replace(idx, idx + lng, "<text:p style-name=\"Default\"/>");

        try {
            FileOutputStream dest = new FileOutputStream(outFile);
            JarOutputStream out = new JarOutputStream(new BufferedOutputStream(dest));
            for(Enumeration<? extends JarEntry> ent = srcODT.entries(); ent.hasMoreElements(); ) {
                JarEntry ze = ent.nextElement();
                out.putNextEntry(new JarEntry(ze.getName()));
                if (ze.getName().equals("content.xml")) {
                    PrintWriter pw = new PrintWriter(out);
                    pw.write(outbuffer.toString());
                    pw.flush();
                }
                else {
                    byte data[] = new byte[2000];
                    int count;
                    BufferedInputStream bis = new BufferedInputStream(srcODT.getInputStream(ze));
                    while ((count = bis.read(data, 0, 2000)) != -1)
                        out.write(data, 0, count);
                    bis.close();
                }
            }
          // Add the plot files...
            for (Iterator<String> i=plotFiles.iterator(); i.hasNext(); ) {
                String pf = i.next();
                out.putNextEntry(new JarEntry("Pictures/" + pf));
                BufferedInputStream bis = new BufferedInputStream(
                    new FileInputStream(pf));
                byte data[] = new byte[2000];
                int count;
                while ((count = bis.read(data, 0, 2000)) != -1)
                    out.write(data, 0, count);
                bis.close();
            }

            out.close();
            parent.message("\nSuccessfully created the result file " + outFile);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    // private routine to get the next <text:p block & set thisLine and thisParTags
    // if accum == true, we add anything in between this and previous paragraph
    // to the buffer -- usually what we want.  Call with accum = false if we're
    // just looking ahead to see what's next
    void getPar(boolean accum) {
        prevEnd = currBeg = currEnd;
        int p = inbuffer.indexOf("<text:p", currEnd);
        if (p < 0) {
            thisLine = thisParTags = null;
            outbuffer.append(inbuffer.substring(prevEnd));
            inbuffer = null;   // toss it; not needed anymore
            return;
        }
        currBeg = p;
        int q = inbuffer.indexOf(">", p);
        if (inbuffer.charAt(q-1) == '/') { // blank line
            currEnd = q + 1;
            thisLine = "";
        }
        else {
            currEnd = inbuffer.indexOf("</text:p>", p);
            thisLine = inbuffer.substring(q+1, currEnd);
            currEnd += 9;
        }
        thisParTags = inbuffer.substring(p, q+1);
        if (accum && currBeg > prevEnd)  // add any intervening stuff to outbuffer
            outbuffer.append(inbuffer.substring(prevEnd, currBeg));
/* omit till we know this is safe
        if (currBeg > 2000) { // garbage collect on inbuffer
            inbuffer.delete(0,currBeg);
            currEnd -= currBeg;
            currBeg = prevEnd = 0;
        }
*/
        parNo++;
    }

    // Usual use of getPar -- accumulate intervening segments
    void getPar() {
        getPar(true);
    }



    // Gather the code lines and save in codeStorage
    // Assumes current line is 1st line of code after header
    // On exit, we are positioned at the next par after the code chunk
    void pullCode() {
        boolean haveCode = true;
        codeStorage = new Vector<String>();
        do {
            int p = thisParTags.indexOf("SWcodebody");
            if (p > 0) {
                codeStorage.addElement(xml2txt(thisLine));
            }
            else {
                p = thisParTags.indexOf("SWcoderef");
                if (p < 0) {
                    haveCode = false;
                }
                else {
                    thisLine = xml2txt(thisLine);
                    p = thisLine.indexOf("{");         // any args?
                    if (p < 0) {                       // no...
                        thisExpr = thisLine;
                        codeRefArgs = null;
                    }
                    else {                             // yes, get 'em
                        thisExpr = thisLine.substring(0,p);
                        Matcher mat = bracePat.matcher(thisLine);
                        codeRefArgs = new Vector<String>(3);
                        while(mat.find())
                            codeRefArgs.addElement(xml2txt(mat.group(1)));
                    }
                    codeStorage.addElement("!coderef " + thisExpr);
                }
            }
            if (haveCode) {    // we've digested this line.  Clear out and get the next one
                getPar(false);
            }
        } while (haveCode);

        if (codeStorage.size() == 0) codeStorage = null;
        codeIter = null;

        currEnd = prevEnd; // go back to previous position
    }

    /**
    * Return the next tag type found in the file
    */
    public Tag nextTag() {
        while (!exprFlag) {
            getPar();
            // now thisParTags has the opening tags, thisLine has the body
            if (thisLine == null)
                return Tag.noMore;
            else if (thisParTags.indexOf("SWcodehead") > 0) {
                thisLang = lastLang = checkColon();  // leaving thisLine as option list
                thisOpts = xml2txt(thisLine);
                getPar();
                pullCode();
                if (codeStorage != null) return Tag.chunk;
                else parent.warning("Warning -- " + getPosition() + ": Empty code chunk");
            }
            else if (thisParTags.indexOf("SWcodebody") > 0) { // get this only if no codehead in source
                thisLang = lastLang;
                thisOpts = "";
                pullCode();
                return Tag.chunk;
            }
            if (thisParTags.indexOf("SWcoderef") > 0) {
                parent.error(getPosition() + ": Code reference outside of a code chunk", Err.SRCERR);
            }
            else if (thisParTags.indexOf("SWopts") > 0) {
                thisLang = checkColon();
                thisOpts = xml2txt(thisLine);
                return (thisLang==null) ? Tag.globalOpts : Tag.langOpts;
            }
            else if (thisParTags.indexOf("SWrecall") > 0) {
                String type = checkColon();    // thisLine should have the label
                if (type.equals("code")) return Tag.recallCode;
                if (type.equals("out") || type.equals("output")) return Tag.recallOutput;
                if (type.equals("fig")) return Tag.recallFig;
                parent.error(getPosition() + ": Cannot recall an element of type \"" + type + "\"", Err.SRCERR);
            }
            else if (thisLine.indexOf("SWexpr") > 0) {
                exprFlag = true;
            }
            else { // just plain old par, nothing at all special
                outbuffer.append(inbuffer.substring(currBeg, currEnd));
            }
        }
    // ---- look at pending line
        if (parent.isWeaving() && exprFlag) {
            // We'll insert bookmarks in thisLine...
            exprBeg = thisLine.indexOf("<text:span text:style-name=\"SWexpr");
            if (exprBeg < 0) {
                String newPar = thisParTags + thisLine + "</text:p>";
                outbuffer.append(newPar);
                exprFlag = false;
                return nextTag();
            }
            else {
                int p = 1 + thisLine.indexOf(">", exprBeg);
                int q = thisLine.indexOf("</text:span", p);
                exprEnd = 1 + thisLine.indexOf(">", q);
                thisExpr = thisLine.substring(p,q);
                q = thisExpr.indexOf(":");
                if (q < 0)
                    thisLang = lastLang;
                else {
                    thisLang = thisExpr.substring(0, q);
                    thisExpr = thisExpr.substring(q+1);
                }
                return Tag.expr;
            }
        }
        return nextTag();
    }



    // Get colon expression at beg of thisLine; lops it off of thisLine
    // i.e., if thisLine is "SAS: fig, !echo  " then we return "SAS"
    // and thisLine becomes "fig, !echo"
    String checkColon() {
        thisLine = xml2txt(thisLine);
        Matcher mat = colonPat.matcher(thisLine);
        if (!mat.find())
            return null; //thisLine.trim();
        // (else)
        String rtn = mat.group(1);
        thisLine = thisLine.substring(mat.end()).trim();
        return rtn;
    }

    /** Read one line from source file (NOT NEEDED since no syntax interface) */
    public String getLine() {
        return null;
    }

    /**
    * Return the current line number in the form <tt>"line #"</tt>.
    */
    public String getPosition() {
        return "par " + parNo;
    }


    // Convert from XML to plain text
    String xml2txt(String stg) {

        // Substitute blanks for <text:s ...> patterns
        String blank16 = "                ";
        Matcher mat = spacePat.matcher(stg);
        StringBuffer sb = new StringBuffer();
        while(mat.find()) {
            int nspc = Integer.valueOf(mat.group(1)),
                lim = (int)Math.floor(nspc/20.0);
            mat.appendReplacement(sb, blank16.substring(0, nspc%16));
            for (int i=0; i<lim; i++) sb.append(blank16);
        }
        mat.appendTail(sb);
        stg = sb.toString();

        // anything left will be formatting tags -- remove them
        stg = stg.replaceAll("<[^>]+>", "");

        // Substitute special XML character markup
        for (int i=0; i<xmlSub.length; i++)
            stg = stg.replaceAll(xmlSub[i][0], xmlSub[i][1]);

        // Take care of auto-format byproducts
        for (int i=0; i<xmlSub1.length; i++)
            stg = stg.replaceAll(xmlSub1[i][0], xmlSub1[i][1]);

        return stg;
    }

    // Convert plain text to appropriate XML markup
    String txt2xml(String stg) {
        // Substitute special XML character markup (reverse map from xml2txt)
        // Have to do it in reverse order too, else we get ampersands!
        for (int i=xmlSub.length-1; i>=0; i--)
            stg = stg.replaceAll(xmlSub[i][1], xmlSub[i][0]);

        // Replace multiple spaces
        StringBuilder sb = new StringBuilder(stg);
        for (int q, p = 0;  (p = sb.indexOf("  ", p)) >= 0;  ) {
            for (q=p; q < sb.length() && sb.charAt(q) == ' '; q++);
            sb.replace(p, q, "<text:s text:c=\"" + (q - p) + "\"/>");
            p = q;
        }
        if (sb.indexOf(" ", 0) == 0)  // initial single space
            sb.replace(0, 1, "<text:s/>");

        return sb.toString();
    }


    // Make an XML paragraph with the given text and style.
    String xmlPar(String txt, String style) {
        String pre = "<text:p text:style-name=\"" + style + "\">",
            post = "</text:p>", line;
        StringBuilder sb = new StringBuilder();
        sb.append(lfSig);  // bookmark for a linefeed
        for(int beg=0, end;  (end=txt.indexOf("\n",beg)) > 0; beg=end+1) {
            line = txt.substring(beg,end);
            sb.append(pre + txt2xml(line) + post);
        }
        sb.append(lfSig);  // bookmark for a linefeed
        return sb.toString();
    }


 /** Return one line from a code chunk.
    <p>If it is actually a code reference, return the string
    <tt>!coderef <i>label</i></tt> or <tt>!coderef <i>*label</i></tt>,
    where <tt><i>label</i></tt> is the label provided.  (Use the <tt>*</tt>
    form to force the code reference to be displayed regardless of
    the value of the option <tt>showref</tt>.)
    <p>Return <tt>null</tt> when the end of the code chunk is reached.
    */
    public String getCodeLine() {
        if (codeIter == null)
            codeIter = codeStorage.iterator();
        return codeIter.hasNext() ? codeIter.next() : null;
    }


    /** Return the language associated with latest tag
    <p> Needed when tag is <tt>chunk</tt>, <tt>langOpts</tt>, or <tt>expr</tt>
    */
    public String getLang() {
        return thisLang;
    }

    /** Return the expression given in the current expr tag */
    public String getExpr() {
        return xml2txt(thisExpr);
    }

    /** Return the options list for latest tag.
    <p> Needed when tag is <tt>chunk</tt>, <tt>globalOpts</tt>, <tt>langOpts</tt>,
    <tt>recallFig</tt>, <tt>recallCode</tt>, or <tt>recallOutput</tt>.
    */
    public String getOptions() {
        return thisOpts;
    }

    /** Return the label specified in the latest tag.
    <p>Needed when the tag is <tt>recallFig</tt>, <tt>recallCode</tt>,
    <tt>recallOutput</tt>.  (When the tag is <tt>chunk</tt>, the label, if any, should be included
    in the list returned by <tt>getOptions()</tt>.) */
    public String getLabel() {
        return thisLine.trim();  // we're using thisLine to save the label
    }

    /** Return the arguments specified in a <tt>coderef</tt> tag. */
    public java.util.Vector<String> getArgs() {
        return codeRefArgs;
    }

    /** Replace the current tagged region of the source file
    with the specified <tt>sig</tt>.  This provides a place-holder for results to be
    incorporated later.
    */
    public void saveBookmark(String sig) {
        if (exprFlag) {
            thisLine = thisLine.substring(0,exprBeg) + sig + thisLine.substring(exprEnd);
        }
        else {
            outbuffer.append(sig);
        }
    }

    /** Find the given bookmark previously saved with signature <tt>sig</tt>, and replace it
    with text.  The <tt>context</tt> argument is a reminder of what kind of content
    is being added.  Implementation may call <tt>StatWeave</tt>'s <tt>getOption()</tt> method
    and expect to retrieve the options in effect for the text being inserted.
    */
    public void replaceBookmark(String sig, String text, Tag context) {
        int beg = outbuffer.indexOf(sig), end = beg + sig.length();
        if (beg < 0) {
            parent.warning("Failed attempt to replace bookmark " + sig);
            return;
        }
        if (context == Tag.recallOutput) {
            if (!"xml".equals(parent.getOption("results")))
                text = xmlPar(text, parent.getOption("outstyle", "Woutput"));
        }
        if (context == Tag.recallCode)
            text = xmlPar(text, parent.getOption("codestyle", "Winput"));
        outbuffer.replace(beg, end, text);
    }


    /** Replace bookmark with specified figure(s) in <tt>ff</tt> */
    public void embedPlots(String sig, FigFile ff, int[] pages) {
        String embedStg =
            "<text:p text:style-name=\"Standard\"><draw:frame  draw:style-name=\"fr2\" draw:name=\"%filename%\" text:anchor-type=\"paragraph\" svg:width=\"%width%in\" svg:height=\"%height%in\" draw:z-index=\"3\"><draw:image xlink:href=\"Pictures/%filename%\" xlink:type=\"simple\" xlink:show=\"embed\" xlink:actuate=\"onLoad\"/></draw:frame></text:p>";
        embedStg = embedStg.replace("%filename%", ff.getFileName());
        double scale = parent.getDim("scale",0),
            wid = parent.getDim("dispw",0), ht = parent.getDim("disph",0);
        if (scale==0) {
            if (wid + ht == 0) { // nothing given -- use orig size
                wid = ff.getWidth();
                ht = ff.getHeight();
            }
            else if(wid > 0) {
                if (ht==0) ht = wid * ff.getHeight() / ff.getWidth();
            }
            else /* wid==0 */
                wid = ht * ff.getWidth() / ff.getHeight();
        }
        else {
            wid = scale * ff.getWidth();
            ht = scale * ff.getHeight();
        }
        embedStg = embedStg.replace("%width%", "" + wid);
        embedStg = embedStg.replace("%height%", "" + ht);
        replaceBookmark(sig, lfSig + embedStg, Tag.recallFig);
        // I didn't put a lfSig afterward because OpenOffice seems to add an extra line
        plotFiles.add(ff.getFileName());
    }


} //----------- END ------------------------------------

